/******************************************************************************
 *
 *    Name: UIDialogLibUtil.c
 *
 *    $Id$
 *
 *    Purpose: A generic implementation of UI Dialog Library utility functions.
 *
 *    Copyright 2013-2023 Citrix Systems, Inc. All rights reserved.
 *
 *****************************************************************************/
#include <string.h>
#include "UIDialogLib.h"
#include "UIDialogLibUtil.h"
#define __USE_GNU /* For RTLD_DEFAULT. */
#include <dlfcn.h>
#include <pthread.h>

static const int REFRESH_TIME_10_MS = 10000; // 10 MILLI SECOND
static const int REFRESH_TIME_2_S = 2000000; // 1 SECOND
#define LOOP_COUNT 10
#define GTK2_VER 2
#define GTK3_VER 3
guint focus_in_event_src;

static pthread_mutex_t  destroy_window_mutex = PTHREAD_MUTEX_INITIALIZER;

/*****************************************************************************/
/*         Implementations on the UIDialogLibUtil external functions.        */
/*****************************************************************************/

/** @brief UIDialogVersion() returns the UIDialogLib major & minor version
 *         numbers. 
 * @param [out] major   The major version number.
 * @param [out] minor   The minor version number.
 */
void UIDialogVersion(unsigned int *major, unsigned int *minor)
{
    *major = UIDIALOG_MAJOR;
    *minor = UIDIALOG_MINOR;
}

/** @brief UIDialogMnemonicAtChar() takes a mnemonicString with marker 
 *  characters indicating the mnemonic and what the mnemonic marker is. 
 *  Both GTK+ and Qt do this, using an '_' and '&' respectively.
 *  The UIDialogMnemonicAtChar() uses this information to strip the mnemonic
 *  markers from the passed in string and return a pointer to the first
 *  mnemonic character.
 * @param mnemonicString    The mnemonicString to process.
 * @param marker            The mnemonic marker character.
 * @return  Returns a pointer to the mnemonic character within the updated
 *          mnemonicString. If no mnemonic is found NULL is returned.
 */
/* UIDialogMnemonicAtChar() searches a given mnemonicString for the mnemonic
   marker character. Both GTK+ and Qt do this, using an '_' and '&'
   respectively. Treat a double marker character as a single marker character.
   This is certainly the case in GTK+ not 100% sure in Qt. */
char* UIDialogMnemonicAtChar(char *mnemonicString, const char marker)
{
    char *mnemonic = NULL;
    char *iter = mnemonicString;
    int length = strlen(mnemonicString) + 1; /* Plus 1 for NULL terminator. */

    /* Keep searching until there is no more string. */
    while (iter && (iter = strchr(iter, marker)) != NULL)
    {
        /* Remove the marker character. */
        memmove(iter, iter + 1, length - (iter - mnemonicString) - 1);

        /* Check if the next character is not another marker. */
        if (mnemonic == NULL && iter[0] != marker)
            mnemonic = iter;

        iter++; /* Skip the next character after a marker. */           
    }

    return mnemonic;
}

void window_realise_func_wrapper(GtkWidget *gtkWidget, gpointer data) {

    window_realise_func = (void (*)(Display*, Window))dlsym(RTLD_DEFAULT, "window_realise_cb");
    if(window_realise_func ==NULL) {
        return;
    }

    pthread_mutex_lock(&destroy_window_mutex);
    Display* xDisplay = NULL;
    Window xWindowId = 0;

    xDisplay = GDK_WINDOW_XDISPLAY(gtk_widget_get_window(GTK_WIDGET(gtkWidget)));
    if (UIDIALOG_MAJOR == GTK3_VER) {
        xWindowId = GDK_WINDOW_XID(gtk_widget_get_window(GTK_WIDGET(gtkWidget)));
    } else if(UIDIALOG_MAJOR == GTK2_VER) {
        xWindowId = GDK_WINDOW_XWINDOW(gtk_widget_get_window(GTK_WIDGET(gtkWidget)));
    }
    window_realise_func(xDisplay, xWindowId);
    pthread_mutex_unlock(&destroy_window_mutex);
}

void *checkWidget(void *ptr) {
    GtkWidget *gtkWidget;
    gtkWidget = (GtkWidget *) ptr;
    int loopVar = 0;
    pthread_mutex_lock(&destroy_window_mutex);
    usleep(REFRESH_TIME_2_S);//2s

     for (loopVar = 0; GTK_IS_WINDOW (gtkWidget) && GTK_IS_WIDGET(gtkWidget) && loopVar < LOOP_COUNT; loopVar++)
     {
         usleep(REFRESH_TIME_10_MS);//10ms
     }
     if (GTK_IS_WINDOW (gtkWidget) && GTK_IS_WIDGET(gtkWidget) && loopVar == LOOP_COUNT) {
        printf("Widget still valid \n");
     }
     window_destroy_handler_func = (void (*)(void))dlsym(RTLD_DEFAULT, "window_destroy_cb");
     if (window_destroy_handler_func != NULL)
        window_destroy_handler_func();
    pthread_mutex_unlock(&destroy_window_mutex);
}

void window_destroy_func_wrapper(GtkWidget *gtkWidget, gpointer data) {
    int rc = 0;
    pthread_t widgetCheckThread;
    rc = pthread_create(&widgetCheckThread, NULL, checkWidget, gtkWidget);
    if (rc != 0) {
       printf("Thread create failed \n");
	}
}


gboolean focus_in_event_callback (GtkWidget* self, GdkEventFocus event, gpointer user_data)
{
    window_focus_in_event_cb_func = (void (*)(Window))dlsym(RTLD_DEFAULT, "focus_in_event_cb");
    if(window_focus_in_event_cb_func ==NULL) {
        return FALSE;
    }
    Window xWindowId = 0;
    if (UIDIALOG_MAJOR == GTK3_VER) {
        xWindowId = GDK_WINDOW_XID(gtk_widget_get_window(GTK_WIDGET(self)));
    } else if(UIDIALOG_MAJOR == GTK2_VER) {
        xWindowId = GDK_WINDOW_XWINDOW(gtk_widget_get_window(GTK_WIDGET(self)));
    }
    window_focus_in_event_cb_func(xWindowId);
    g_signal_handler_disconnect (self, focus_in_event_src);
    return FALSE;
}
